from datetime import datetime
from typing import Literal, TypeVar

import numpy as np
from pymatgen.io.validation.check_kpoints_kspacing import (
    get_kpoint_divisions_from_kspacing,
)

from emmet.builders.base import BaseBuilderInput
from emmet.builders.utils import filter_map
from emmet.core.material import PropertyOrigin
from emmet.core.math import Matrix3D, Vector6D
from emmet.core.polar import DielectricDoc, PiezoelectricDoc
from emmet.core.tasks import CoreTaskDoc
from emmet.core.types.typing import MaterialIdentifierType, NullableDateTimeType

PIEZO_TENSOR_TYPE = tuple[Vector6D, Vector6D, Vector6D]


class BaseLinearResponseInput(BaseBuilderInput):
    last_updated: NullableDateTimeType
    origins: list[PropertyOrigin]


class DielectricBuilderInput(BaseLinearResponseInput):
    epsilon_static: Matrix3D
    epsilon_ionic: Matrix3D


class PiezoelectricBuilderInput(BaseLinearResponseInput):
    piezo_static: PIEZO_TENSOR_TYPE
    piezo_ionic: PIEZO_TENSOR_TYPE


def build_dielectric_docs(
    linear_resp_input: list[DielectricBuilderInput], **kwargs
) -> list[DielectricDoc]:
    return list(
        filter_map(
            DielectricDoc.from_ionic_and_electronic,
            linear_resp_input,
            work_keys=[
                "deprecated",
                "material_id",
                "structure",
                "origins",
                "epsilon_static",
                "epsilon_ionic",
                "last_updated",
            ],
            **kwargs,
        )
    )


def build_piezo_docs(
    linear_resp_input: list[PiezoelectricBuilderInput],
    **kwargs,
) -> list[PiezoelectricDoc]:
    return list(
        filter_map(
            PiezoelectricDoc.from_ionic_and_electronic,
            linear_resp_input,
            work_keys=[
                "deprecated",
                "material_id",
                "structure",
                "origins",
                "piezo_static",
                "piezo_ionic",
                "last_updated",
            ],
            **kwargs,
        )
    )


# -----------------------------------------------------------------------------
# Helper funcs + types
# -----------------------------------------------------------------------------

CENTROSYMMETRIC_SPACE_GROUPS = [
    "-1",
    "2/m",
    "mmm",
    "4/m",
    "4/mmm",
    "-3",
    "-3m",
    "6/m",
    "6/mmm",
    "m-3",
    "m-3m",
]


def filter_piezo_tasks(
    tasks: list[CoreTaskDoc],
) -> list[CoreTaskDoc | None]:
    """
    Yields list of CoreTaskDocs with spacegroups appropriate for
    ``build_piezo_docs``

    Can be used to filter input ``tasks`` for ``obtain_blessed_linear_builder_input``
    """
    return list(
        filter(
            lambda x: x.input.structure.get_space_group_info()[0]
            in CENTROSYMMETRIC_SPACE_GROUPS,
            tasks,
        )
    )


T = TypeVar("T", DielectricBuilderInput, PiezoelectricBuilderInput)


def obtain_blessed_linear_builder_input(
    tasks: list[CoreTaskDoc],
    property_name: Literal["dielectric", "piezoelectric"],
    target: T,
    material_id: MaterialIdentifierType | None = None,
    material_last_updated: datetime | None = None,
) -> T:
    """
    Yield a target document [``DielectricBuilderInput | PiezoelectricBuilderInput``]
    from a list of CoreTaskDocs using the 'best' document in the list.

    [Optional] Relevant material properties needed if running in context of
    building properties for 'material's:
       - material_id -> anchor identifier for all ``tasks``
       - material_last_updated -> when anchor material document was last updated

    Relevant CoreTaskDoc fields needed:
        - task_id
        - last_updated
        - input.is_hubbard
        - input.kpoints
        - input.structure
        - orig_inputs.kpoints
        - output.outcar
        - output.bandgap

    CoreTaskDocs need to have non-null bandgap
    """
    relevant_tasks = [
        {
            "structure": task.input.structure,
            "task_id": task.task_id,
            "nkpoints": _parse_kpoints(task),
            **{
                k: task.output.outcar.get(k)
                for k in (
                    "epsilon_static",
                    "epsilon_ionic",
                    "piezo_static",
                    "piezo_ionic",
                )
            },
            "is_hubbard": int(task.input.is_hubbard),
            "task_last_updated": task.last_updated,
        }
        for task in tasks
        if task.output.bandgap > 0.0
    ]

    best_task = sorted(
        relevant_tasks,
        key=lambda doc: (
            # Entries with any None sort last (False < True, reversed)
            doc["is_hubbard"] is not None
            and doc["nkpoints"] is not None
            and doc["task_last_updated"] is not None,
            doc["is_hubbard"],
            doc["nkpoints"],
            doc["task_last_updated"],
        ),
        reverse=True,
    )[0]

    doc = {
        "material_id": material_id,
        "last_updated": material_last_updated,
        "origins": [
            PropertyOrigin(
                name=property_name,
                task_id=best_task["task_id"],
                last_updated=best_task["task_last_updated"],
            )
        ],
        **best_task,
    }

    return target(**doc)


def _parse_kpoints(task: CoreTaskDoc) -> int:

    for inp_field in ("input", "orig_inputs"):
        if (
            kpts := getattr(getattr(task, inp_field, None), "kpoints", None)
        ) is not None:
            break

    if kpts is None:
        if isinstance(dk := (task.input.incar or {}).get("KSPACING"), float):
            return np.prod(get_kpoint_divisions_from_kspacing(task.structure, dk))
        return 0

    if kpts.style.name in ("Monkhorst", "Gamma"):
        return np.prod(kpts.kpts[0])
    return task.orig_inputs.kpoints.num_kpts



import logging
import warnings

from pydantic import BaseModel, Field
from pymatgen.analysis.phase_diagram import PhaseDiagram, PhaseDiagramError
from pymatgen.entries.computed_entries import ComputedStructureEntry

from emmet.builders.utils import HiddenPrints
from emmet.core.thermo import PhaseDiagramDoc, ThermoDoc
from emmet.core.types.enums import ThermoType
from emmet.core.types.pymatgen_types.computed_entries_adapter import (
    ComputedStructureEntryType,
)
from emmet.core.vasp.calc_types.enums import RunType


class ThermoBuilderInput(BaseModel):
    """
    Minimum inputs required to build ThermoDocs and PhaseDiagramDocs
    for a chemical system.
    """

    chemsys: str = Field(
        ...,
        description="Dash-delimited string of elements in the material.",
    )

    entries: dict[RunType | ThermoType, list[ComputedStructureEntryType]] = Field(
        ...,
        description="""
        List of all computed entries for 'chemsys' that are valid for the specified thermo type.
        Entries for elemental endpoints of 'chemsys' are required.
        """,
    )


class ThermoBuilderOutput(BaseModel):
    """Output of build_thermo_docs_and_phase_diagram_docs function"""

    chemsys: str
    thermo_docs: dict[RunType | ThermoType, list[ThermoDoc] | None]
    phase_diagram_docs: dict[RunType | ThermoType, PhaseDiagramDoc | None]


ThermoPDPair = tuple[list[ThermoDoc] | None, PhaseDiagramDoc | None]

logger = logging.getLogger()


def build_thermo_docs_and_phase_diagram_docs(
    thermo_input: ThermoBuilderInput,
) -> ThermoBuilderOutput:
    chemsys = thermo_input.chemsys

    thermo_docs = dict()
    phase_diagram_docs = dict()
    for thermo_type, entry_list in thermo_input.entries.items():
        logger.debug(
            f"Processing {len(entry_list)} entries for: {chemsys} and thermo type: {thermo_type}"
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with HiddenPrints():
                _thermo_docs, _phase_diagram_doc = _produce_pair(
                    entry_list, thermo_type
                )
                thermo_docs[thermo_type] = _thermo_docs
                phase_diagram_docs[thermo_type] = _phase_diagram_doc

    return ThermoBuilderOutput(
        chemsys=chemsys,
        thermo_docs=thermo_docs,
        phase_diagram_docs=phase_diagram_docs,
    )


def _produce_pair(
    computed_structure_entries: list[ComputedStructureEntry],
    thermo_type: RunType | ThermoType,
) -> ThermoPDPair:
    phase_diagram_doc = None
    try:
        phase_diagram: PhaseDiagram = ThermoDoc.construct_phase_diagram(
            computed_structure_entries
        )
        thermo_docs: list[ThermoDoc] = ThermoDoc.from_entries(
            computed_structure_entries,
            thermo_type,
            phase_diagram,
            use_max_chemsys=True,
            deprecated=False,
        )

        if phase_diagram:
            chemsys = "-".join(
                sorted(set([el.symbol for el in phase_diagram.elements]))
            )
            phase_diagram_id = f"{chemsys}_{thermo_type.value}"
            phase_diagram_doc = PhaseDiagramDoc(
                phase_diagram_id=phase_diagram_id,
                chemsys=chemsys,
                phase_diagram=phase_diagram,
                thermo_type=thermo_type,
            )

        if not thermo_docs:
            return None, phase_diagram_doc

        return thermo_docs, phase_diagram_doc

    except PhaseDiagramError as p:
        elsyms = []
        for entry in computed_structure_entries:
            elsyms.extend([el.symbol for el in entry.composition.elements])

        logger.error(
            f"Phase diagram error in chemsys {'-'.join(sorted(set(elsyms)))}: {p}"
        )

        return None, None



import numpy as np

from emmet.builders.base import BaseBuilderInput
from emmet.core.similarity import (
    CrystalNNSimilarity,
    M3GNetSimilarity,
    SimilarityDoc,
    SimilarityEntry,
    SimilarityMethod,
)

SIM_METHOD_TO_SCORER = {
    SimilarityMethod(k): v
    for k, v in {
        "CrystalNN": CrystalNNSimilarity,
        "M3GNet": M3GNetSimilarity,
    }.items()
}


class SimilarityBuilderInput(BaseBuilderInput):
    """Augment base builder input with extra fields."""

    similarity_method: SimilarityMethod
    feature_vector: list[float]


# this could probably be parallelized over `similarity_method`
def build_feature_vectors(
    input_documents: list[BaseBuilderInput],
    similarity_method: SimilarityMethod | str = SimilarityMethod.CRYSTALNN,
) -> list[SimilarityBuilderInput]:
    """Generate similarity feature vectors.

    Args:
        input_documents : list of BaseBuilderInput to process
        similarity_method : SimilarityMethod = SimilarityMethod.CRYSTALNN
            The method to use in building similarity docs.
    Returns:
        list of SimilarityBuilderInput
    """
    if isinstance(similarity_method, str):
        similarity_method = (
            SimilarityMethod[similarity_method]
            if similarity_method in SimilarityMethod.__members__
            else SimilarityMethod(similarity_method)
        )

    if scorer_cls := SIM_METHOD_TO_SCORER.get(similarity_method):
        scorer = scorer_cls()
    else:
        raise ValueError(f"Unsupported {similarity_method=}")

    return list(
        map(
            lambda x: SimilarityBuilderInput(
                material_id=x.material_id,
                structure=x.structure,
                similarity_method=similarity_method,
                feature_vector=scorer._featurize_structure(x.structure),
            ),
            input_documents,
        )
    )


def build_similarity_docs(
    input_documents: list[SimilarityBuilderInput],
    num_closest: int = 100,
) -> list[SimilarityDoc]:
    """Generate similarity feature vectors.

    All input docs should use the same similarity method.
    A check is performed at the start to ensure this.

    Args:
        input_documents : list of SimilarityBuilderInput to process
        num_closest : int = 100
            The number of most similar materials to identify
            for each material
    Returns:
        list of SimilarityDoc
    """

    if (
        len(distinct_sim_methods := {doc.similarity_method for doc in input_documents})
        > 1
    ):
        raise ValueError(
            f"Multiple similarity methods found: {', '.join(distinct_sim_methods)}"
        )

    scorer_cls = SIM_METHOD_TO_SCORER[method := input_documents[0].similarity_method]
    material_ids, vectors, structures = np.array(
        [doc.material_id, doc.feature_vector, doc.structure] for doc in input_documents
    ).T

    similarity_docs = []
    for i, material_id in enumerate(material_ids):
        closest_idxs, closest_dist = scorer_cls._get_closest_vectors(
            i, vectors, num_closest
        )
        similarity_docs.append(
            SimilarityDoc.from_structure(
                meta_structure=structures[i],
                material_id=material_id,
                feature_vector=vectors[i],
                method=method,
                sim=[
                    SimilarityEntry(
                        task_id=material_ids[jdx],
                        nelements=len(structures[jdx].composition.elements),
                        dissimilarity=100.0 - closest_dist[j],
                        formula=structures[jdx].formula,
                    )
                    for j, jdx in enumerate(closest_idxs)
                ],
            )
        )
    return similarity_docs



from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from emmet.builders.base import BaseBuilderInput
from emmet.builders.utils import filter_map, try_call
from emmet.core.bonds import BondingDoc


def build_bonding_docs(
    input_documents: list[BaseBuilderInput], **kwargs
) -> list[BondingDoc]:
    """
    Generate bonding documents from input structures.

    Transforms a list of BaseBuilderInput documents containing
    Pymatgen structures into corresponding BondingDoc instances by
    analyzing the bonding environment of each structure.

    Caller is responsible for creating BaseBuilderInput instances
    within their data pipeline context.

    Args:
        input_documents: List of BaseBuilderInput documents to process.

    Returns:
       list[BondingDoc]
    """

    def _build(deprecated: bool, material_id: str, structure, **kwargs) -> BondingDoc:
        return BondingDoc.from_structure(
            deprecated=deprecated,
            material_id=material_id,
            structure=try_call(
                lambda s: SpacegroupAnalyzer(s).get_conventional_standard_structure(),
                structure,
            ),
            **kwargs
        )

    return list(
        filter_map(
            _build,
            input_documents,
            work_keys=["deprecated", "material_id", "structure"],
            **kwargs
        )
    )



from robocrys import __version__ as __robocrys_version__
from robocrys.condense.mineral import MineralMatcher

from emmet.builders.base import BaseBuilderInput
from emmet.builders.utils import filter_map
from emmet.core.robocrys import RobocrystallogapherDoc


def build_robocrys_docs(
    input_documents: list[BaseBuilderInput], **kwargs
) -> list[RobocrystallogapherDoc]:
    """
    Generate robocrystallographer descriptions from input structures.

    Transforms a list of BaseBuilderInput documents containing
    Pymatgen structures into corresponding RobocrystallogapherDoc instances by
    using robocrys' StructureCondenser and StructureDescriber classes.

    Caller is responsible for creating BaseBuilderInput instances
    within their data pipeline context.

    Args:
        input_documents: List of BaseBuilderInput documents to process.

    Returns:
        list[RobocrystallogapherDoc]
    """
    mineral_matcher = MineralMatcher()
    return list(
        filter_map(
            RobocrystallogapherDoc.from_structure,
            input_documents,
            work_keys=["deprecated", "material_id", "structure", "origins"],
            mineral_matcher=mineral_matcher,
            robocrys_version=__robocrys_version__,
            **kwargs
        )
    )



from pydantic import BaseModel

from emmet.builders.utils import filter_map
from emmet.core.summary import (
    AbsorptionData,
    BandstructureSummary,
    ChargeDensityData,
    ChemenvData,
    DielectricSummary,
    DosSummary,
    ElasticitySummary,
    ElectrodesData,
    ElectronicStructureSummary,
    EosData,
    GBSummary,
    MagnetismSummary,
    MaterialsSummary,
    OxiStatesSummary,
    PhononData,
    PiezoelectricSummary,
    ProvenanceSummary,
    SubstratesData,
    SummaryDoc,
    SurfacesSummary,
    ThermoSummary,
    XASSummary,
)


class SummaryBuilderInputs(BaseModel):
    """
    Input model for building summary documents.

    Bundles the property summary documents and property shim documents
    needed to construct a single SummaryDoc. Property summary documents
    contribute field values to the resulting SummaryDoc, while property
    shim documents are used solely to populate the has_props mapping.
    """

    property_summary_docs: list[
        MaterialsSummary
        | ThermoSummary
        | XASSummary
        | GBSummary
        | ElectronicStructureSummary
        | BandstructureSummary
        | DosSummary
        | MagnetismSummary
        | ElasticitySummary
        | DielectricSummary
        | PiezoelectricSummary
        | SurfacesSummary
        | OxiStatesSummary
        | ProvenanceSummary
    ]
    property_shim_docs: list[
        ChargeDensityData
        | EosData
        | PhononData
        | AbsorptionData
        | ElectrodesData
        | SubstratesData
        | ChemenvData
    ]


def build_summary_docs(
    input_documents: list[SummaryBuilderInputs], **kwargs
) -> list[SummaryDoc]:
    """
    Generate summary documents from input property documents.

    Transforms a list of SummaryBuilderInputs into corresponding
    SummaryDoc instances by merging property summary documents and
    property shim documents for each material. Each SummaryDoc
    aggregates fields from all provided property summaries and tracks
    which properties are available via the has_props mapping.

    Caller is responsible for creating SummaryBuilderInputs instances
    within their data pipeline context.

    Args:
        input_documents: List of SummaryBuilderInputs documents to process.

    Returns:
        list[SummaryDoc]
    """
    return filter_map(
        SummaryDoc.from_docs,
        input_documents,
        work_keys=["property_summary_docs", "property_shim_docs"],
        **kwargs
    )



from emmet.builders.base import BaseBuilderInput
from emmet.builders.utils import filter_map
from emmet.core.chemenv import ChemEnvDoc


def build_chemenv_docs(
    input_documents: list[BaseBuilderInput], **kwargs
) -> list[ChemEnvDoc]:
    """
    Generate chemical environment documents from input structures.

    Transforms a list of BaseBuilderInput documents containing
    Pymatgen structures into corresponding ChemEnvDoc instances by
    analyzing the chemical environment of each structure.

    Caller is responsible for creating BaseBuilderInput instances
    within their data pipeline context.

    Args:
        input_documents: List of BaseBuilderInput documents to process.

    Returns:
        list[ChemEnvDoc]
    """

    return list(
        filter_map(
            ChemEnvDoc.from_structure,
            input_documents,
            work_keys=["deprecated", "material_id", "structure", "builder_meta"],
            **kwargs
        )
    )



from typing import Any

from emmet.builders.settings import EmmetBuildSettings
from emmet.builders.utils import filter_map, get_potcar_stats
from emmet.core.tasks import CoreTaskDoc, TaskDoc
from emmet.core.vasp.task_valid import TaskDocument
from emmet.core.vasp.validation_legacy import ValidationDoc


def build_validation_doc(
    input_documents: list[CoreTaskDoc | TaskDoc | TaskDocument],
    settings: EmmetBuildSettings = EmmetBuildSettings(),
    potcar_stats: dict[str, Any] = get_potcar_stats(method="stored"),
    **kwargs
) -> list[ValidationDoc]:
    """
    Build a ValidationDoc from a CoreTaskDoc by checking CoreTaskDoc
    parameters against reference values.

    Args:
        input: List of parsed task document to validate.
        settings: Reference values used in validation, defaults defined in EmmetBuildSettings.
            Relevant settings: VASP_KSPACING_TOLERANCE, VASP_DEFAULT_INPUT_SETS, VASP_CHECKED_LDAU_FIELDS,
            VASP_MAX_SCF_GRADIENT, and DEPRECATED_TAGS.
        potcar_stats: POTCAR stats used to validate POTCARs used for the source calculation
            for 'input'. Defaults to compiled values in 'emmet.builders.vasp.mp_potcar_stats.json.gz'

    Returns:
        list[ValidationDoc]
    """
    return list(
        filter_map(
            ValidationDoc.from_task_doc,
            input_documents,
            kspacing_tolerance=settings.VASP_KSPACING_TOLERANCE,
            input_sets=settings.VASP_DEFAULT_INPUT_SETS,
            LDAU_fields=settings.VASP_CHECKED_LDAU_FIELDS,
            max_allowed_scf_gradient=settings.VASP_MAX_SCF_GRADIENT,
            potcar_stats=potcar_stats,
            bad_tags=settings.DEPRECATED_TAGS,
            **kwargs
        )
    )



from emmet.builders.base import BaseBuilderInput
from emmet.builders.utils import filter_map
from emmet.core.absorption import AbsorptionDoc
from emmet.core.material import PropertyOrigin
from emmet.core.types.typing import DateTimeType


class AbsorptionBuilderInput(BaseBuilderInput):
    energies: list[float]
    real_d: list[float]
    imag_d: list[float]
    absorption_co: list[float]
    bandgap: float | None
    nkpoints: int | None
    last_updated: DateTimeType
    origins: list[PropertyOrigin]


def build_absorption_docs(
    input_documents: list[AbsorptionBuilderInput], **kwargs
) -> list[AbsorptionDoc]:
    """
    Generate absorption documents from input structures.

    Transforms a list of AbsorptionBuilderInput documents containing
    Pymatgen structures into corresponding AbsorbtionDoc instances by
    generating an absorption spectrum based on frequency dependent
    dielectric function outputs.

    Caller is responsible for creating AbsorptionBuilderInput instances
    within their data pipeline context.

    Args:
        input_documents: List of AbsorptionBuilderInput documents to process.

    Returns:
       list[AbsorbtionDoc]
    """
    return list(
        filter_map(
            AbsorptionDoc.from_structure,
            input_documents,
            work_keys=[
                "energies",
                "real_d",
                "imag_d",
                "absorption_co",
                "bandgap",
                "nkpoints",
                "last_updated",
                "origins",
                # PropertyDoc.from_structure(...) kwargs
                "deprecated",
                "material_id",
                "structure",
            ],
            **kwargs
        )
    )



"""
Settings for defaults in the build pipelines for the Materials Project
"""

from pydantic.fields import Field

from emmet.core.provenance import Author, History
from emmet.core.qchem.calc_types import TaskType as QChemTaskType
from emmet.core.settings import EmmetSettings
from emmet.core.vasp.calc_types import TaskType as VaspTaskType


class EmmetBuildSettings(EmmetSettings):
    """
    Settings for the emmet-builder module
    The default way to modify these is to modify ~/.emmet.json or set the environment variable
    EMMET_CONFIG_FILE to point to the json with emmet settings
    """

    BUILD_TAGS: list[str] = Field(
        [], description="Tags for calculations to build materials"
    )
    EXCLUDED_TAGS: list[str] = Field(
        [],
        description="Tags to exclude from materials",
    )

    DEPRECATED_TAGS: list[str] = Field(
        [], description="Tags for calculations to deprecate"
    )

    NON_COMMERCIAL_TAGS: list[str] = Field(
        [], description="Tages for which to add BY-NC as license data in builder_meta"
    )

    VASP_ALLOWED_VASP_TYPES: list[VaspTaskType] = Field(
        list(VaspTaskType),
        description="Allowed task_types to build materials from",
    )

    QCHEM_ALLOWED_TASK_TYPES: list[QChemTaskType] = Field(
        list(
            map(
                QChemTaskType,
                [
                    "Single Point",
                    "Force",
                    "Geometry Optimization",
                    "Frequency Analysis",
                    "Frequency Flattening Geometry Optimization",
                ],
            )
        ),
        description="Allowed task_types to build molecules from",
    )

    DEFAULT_REFERENCE: str = Field(
        "@article{Jain2013,\nauthor = {Jain, Anubhav and Ong, Shyue Ping and "
        "Hautier, Geoffroy and Chen, Wei and Richards, William Davidson and "
        "Dacek, Stephen and Cholia, Shreyas and Gunter, Dan and Skinner, David "
        "and Ceder, Gerbrand and Persson, Kristin a.},\n"
        "doi = {10.1063/1.4812323},\nissn = {2166532X},\n"
        "journal = {APL Materials},\nnumber = {1},\npages = {011002},\n"
        "title = {{The Materials Project: A materials genome approach to "
        "accelerating materials innovation}},\n"
        "url = {http://link.aip.org/link/AMPADS/v1/i1/p011002/s1\\&Agg=doi},\n"
        "volume = {1},\nyear = {2013}\n}\n\n@misc{MaterialsProject,\n"
        "title = {{Materials Project}},\nurl = {http://www.materialsproject.org}\n}",
        description="Default bibtex citation for all provenance",
    )

    DEFAULT_AUTHOR: Author = Field(
        Author(name="Materials Project", email="feedback@materialsproject.org"),
        description="Default Author for provenance ",
    )

    DEFAULT_HISTORY: History = Field(
        History(  # type: ignore[call-arg]
            name="Materials Project Optimized Structure",
            url="http://www.materialsproject.org",
        ),
        description="Default History for provenance ",
    )



"""Build provenance collection."""

import logging

from pymatgen.analysis.structure_matcher import ElementComparator, StructureMatcher

from emmet.builders.base import BaseBuilderInput
from emmet.builders.settings import EmmetBuildSettings
from emmet.builders.utils import filter_map
from emmet.core.connectors.analysis import parse_cif
from emmet.core.connectors.icsd.client import IcsdClient
from emmet.core.connectors.icsd.enums import IcsdSubset
from emmet.core.provenance import DatabaseSNL, ProvenanceDoc

SETTINGS = EmmetBuildSettings()
structure_matcher = StructureMatcher(
    ltol=SETTINGS.LTOL,
    stol=SETTINGS.STOL,
    comparator=ElementComparator(),
    angle_tol=SETTINGS.ANGLE_TOL,
    primitive_cell=True,
    scale=True,
    attempt_supercell=False,
    allow_subset=False,
)


logger = logging.getLogger(__name__)


def _get_snl_from_cif(cif_str: str, **kwargs) -> DatabaseSNL | None:
    """Build a database SNL from a CIF plus its metadata.

    NB: Only takes the first structure from a CIF.
    While a CIF can technically contain many structures,
    the ICSD usually only distributes CIFs with one structure
    per file.

    Parameters
    -----------
    cif_str : the CIF to parse
    **kwargs to pass to `DatabaseSNL`
    """
    try:
        structures, cif_parsing_remarks = parse_cif(cif_str)
        remarks = kwargs.pop("remarks", None) or cif_parsing_remarks or None
        snl = DatabaseSNL.from_structure(
            meta_structure=structures[0],
            structure=structures[0],
            **kwargs,
        )

        snl.about.remarks = remarks

    except Exception as e:
        logger.warning(e)
        snl = None

    return snl


def update_experimental_icsd_structures(**client_kwargs) -> list[DatabaseSNL]:
    """Update the collection of ICSD SNLs.

    Parameters
    -----------
    **client_kwargs to pass to `IcsdClient`

    Returns
    -----------
    List of DatabaseSNL
    """
    data = []
    with IcsdClient(use_document_model=False, **client_kwargs) as client:
        for icsd_subset in (
            IcsdSubset.EXPERIMENTAL_METALORGANIC,
            IcsdSubset.EXPERIMENTAL_INORGANIC,
        ):
            data += client.search(
                subset=IcsdSubset.EXPERIMENTAL_INORGANIC,
                space_group_number=(1, 230),
                include_cif=True,
                include_metadata=False,
            )

    parsed = [
        _get_snl_from_cif(
            doc["cif"],
            snl_id=f"icsd-{doc['collection_code']}",
            tags=[doc["subset"].value],
            source="icsd",
        )
        for doc in data
    ]

    return sorted(
        [doc for doc in parsed if doc],
        key=lambda doc: int(doc.snl_id.split("-", 1)[-1]),
    )


def match_against_snls(
    input_doc: BaseBuilderInput,
    snls: list[DatabaseSNL],
):
    """Match a single document against the SNL collection."""
    database_ids = {}
    authors = [SETTINGS.DEFAULT_AUTHOR]
    history = [SETTINGS.DEFAULT_HISTORY]
    references = [SETTINGS.DEFAULT_REFERENCE]
    theoretical = True

    for snl in [
        doc
        for doc in snls
        if doc.chemsys
        == (
            "-".join(sorted(input_doc.structure.composition.chemical_system.split("-")))
        )
    ]:
        if structure_matcher.fit(input_doc.structure, snl.structure):

            if snl.source and snl.source in {"icsd", "pauling"}:
                theoretical = False
                database_ids[snl.source].append(snl.snl_id)

            if snl.about:
                authors.extend(snl.about.authors or [])
                history.extend(snl.about.history or [])
                # `SNLAbout` uses string for `references`,
                # `ProvenanceDoc` uses list of str
                if snl.about.references:
                    references.append(snl.about.references)

    return ProvenanceDoc.from_structure(
        meta_structure=input_doc.structure,
        material_id=input_doc.material_id,
        database_IDs=database_ids,
        theoretical=theoretical,
        authors=authors,
        history=history,
        references=references,
    )


def build_provenance_docs(
    input_documents: list[BaseBuilderInput], snls: list[DatabaseSNL], **kwargs
) -> list[ProvenanceDoc]:
    """Build the provenance collection."""

    return list(filter_map(match_against_snls, input_documents, snls=snls, **kwargs))



from emmet.builders.base import BaseBuilderInput
from emmet.builders.utils import filter_map
from emmet.core.oxidation_states import OxidationStateDoc


def build_oxidation_states_docs(
    input_documents: list[BaseBuilderInput], **kwargs
) -> list[OxidationStateDoc]:
    """
    Generate oxidation state documents from input structures.

    Transforms a list of BaseBuilderInput documents containing
    Pymatgen structures into corresponding OxidationStateDoc instances by
    analyzing the oxidation states of each structure.

    Caller is responsible for creating BaseBuilderInput instances
    within their data pipeline context.

    Args:
        input_documents: List of BaseBuilderInput documents to process.

    Returns:
        list[OxidationStateDoc]
    """
    return list(
        filter_map(
            OxidationStateDoc.from_structure,
            input_documents,
            work_keys=[
                "deprecated",
                "material_id",
                "structure",
                "builder_meta",
            ],
            **kwargs
        )
    )



from itertools import groupby
from typing import Iterator

from emmet.builders.settings import EmmetBuildSettings
from emmet.core.tasks import CoreTaskDoc
from emmet.core.utils import group_structures, undeform_structure
from emmet.core.vasp.calc_types import TaskType
from emmet.core.vasp.material import MaterialsDoc
from pydantic import Field


class ValidationTaskDoc(CoreTaskDoc):
    """
    Wrapper for TaskDoc to ensure compatiblity with validation checks
    in MaterialsDoc.from_tasks(...) if validation builder is skipped
    """

    is_valid: bool = Field(True)


def build_material_docs(
    input_documents: list[ValidationTaskDoc],
    settings: EmmetBuildSettings = EmmetBuildSettings(),
) -> list[MaterialsDoc]:
    """
    Aggregate ValidationTaskDocs into MaterialsDocs by chemical formula.
    Caller is responsible for creating ValidationTaskDoc instances within
    their data pipeline context.

    Groups input documents by formula_pretty, performs structure matching
    on each formula group, and constructs a MaterialsDoc for each group of
    task documents with matching structures within each formula group.

    Args:
        input_documents: List of ValidationTaskDoc objects to process. Must contain
            ALL documents for each unique formula_pretty value to avoid incorrect
            material splitting. Documents for the same formula should not be split
            across multiple function calls.
        settings: Builder configuration settings, defaults defined in EmmetBuildSettings.
            Relevant settings: VASP_STRUCTURE_QUALITY_SCORES, VASP_USE_STATICS,
            VASP_ALLOWED_VASP_TYPES, LTOL, STOL, ANGLE_TOL, and SYMPREC.

    Returns:
        list[MaterialsDoc]
    """

    input_documents.sort(key=lambda x: x.formula_pretty)
    materials = []
    for _, group in groupby(input_documents, key=lambda x: x.formula_pretty):
        # TODO: logging - task_ids = [task.task_id for task in group]
        group = list(group)
        task_transformations = [task.transformations for task in group]
        grouped_tasks = filter_and_group_tasks(group, task_transformations, settings)
        for group in grouped_tasks:
            try:
                doc = MaterialsDoc.from_tasks(
                    group,
                    structure_quality_scores=settings.VASP_STRUCTURE_QUALITY_SCORES,
                    use_statics=settings.VASP_USE_STATICS,
                )
                materials.append(doc)
            except Exception as e:
                # TODO: logging - failed_ids = list({t_.task_id for t_ in group})
                doc = MaterialsDoc.construct_deprecated_material(group)
                doc.warnings.append(str(e))
                materials.append(doc)

    return materials


def filter_and_group_tasks(
    tasks: list[ValidationTaskDoc],
    task_transformations: list[dict | None],
    settings: EmmetBuildSettings,
) -> Iterator[list[ValidationTaskDoc]]:
    """Groups tasks by structure matching"""

    filtered_tasks = []
    filtered_transformations = []
    for task, transformations in zip(tasks, task_transformations):
        if any(
            allowed_type == task.task_type
            for allowed_type in settings.VASP_ALLOWED_VASP_TYPES
        ):
            filtered_tasks.append(task)
            filtered_transformations.append(transformations)
    structures = []
    for idx, (task, transformations) in enumerate(
        zip(filtered_tasks, filtered_transformations)
    ):
        if task.task_type == TaskType.Deformation:
            if transformations is None:
                # Do not include deformed tasks without transformation information
                continue
            else:
                s = undeform_structure(task.input.structure, transformations)
        else:
            s = task.output.structure

        s.index = idx
        structures.append(s)

    grouped_structures = group_structures(
        structures,
        ltol=settings.LTOL,
        stol=settings.STOL,
        angle_tol=settings.ANGLE_TOL,
        symprec=settings.SYMPREC,
    )
    for group in grouped_structures:
        grouped_tasks = [filtered_tasks[struct.index] for struct in group]
        yield grouped_tasks



from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto
from functools import update_wrapper

from pydantic import BaseModel
from pymatgen.analysis.magnetism.analyzer import Ordering

from emmet.builders.utils import filter_map
from emmet.core.electronic_structure import ElectronicStructureDoc
from emmet.core.material import PropertyOrigin
from emmet.core.types.electronic_structure import BSShim, DosShim
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.types.typing import IdentifierType


class InputMeta(BaseModel):
    # structure metadata
    deprecated: bool = False
    material_id: IdentifierType | None = None
    meta_structure: StructureType
    origins: list[PropertyOrigin] = []
    warnings: list[str] = []


class StructureInputs(InputMeta):
    # summary electronic structure data from VASP outputs for task doc
    band_gap: float
    cbm: float | None = None
    vbm: float | None = None
    efermi: float | None = None
    is_gap_direct: bool
    is_metal: bool
    magnetic_ordering: Ordering


class StructuresShim(InputMeta):
    # map of structures with task_ids -> used in post doc build checks
    structures: dict[IdentifierType, StructureType]


class BSInputs(StructuresShim):
    bandstructures: BSShim


class DosInputs(StructuresShim):
    dos: DosShim
    is_gap_direct: bool


class BSDosInputs(DosInputs, BSInputs): ...


class Variant(Enum):
    STRUCTURE = auto()
    BS = auto()
    DOS = auto()
    BS_DOS = auto()


InputData = list[StructureInputs] | list[BSInputs] | list[DosInputs] | list[BSDosInputs]
"""Tagged union for valid input types for build_electronic_structure_docs."""


@dataclass
class ESBuilderInput:
    """
    Container for electronic structure builder inputs that pairs
    a Variant tag with the corresponding input data.

    The variant field determines which construction path
    build_electronic_structure_docs will use to produce
    ElectronicStructureDoc instances.

    The data field holds a list of the appropriate input model
    (StructureInputs, BSInputs, DosInputs, or BSDosInputs) matching
    the chosen variant. Callers are responsible for populating this
    list within their own data pipeline context; helper functions
    such as ``obtain_blessed_dos`` and ``obtain_blessed_bs`` can assist
    in selecting the best candidate calculations for inclusion.
    """

    variant: Variant
    data: InputData


def variant_dispatch(func):
    """
    Slight remix to functools.singledispatch to perform dynamic
    dispatch based on the type of an enum variant for arg, rather
    than arg itself.

    Only usable with objects with a ``variant`` enum attr. See: ESBuilderInput
    """
    registry = {}

    def register(variant_value):
        def decorator(f):
            registry[variant_value] = f
            return f

        return decorator

    def wrapper(inputs, **kwargs):
        variant = inputs.variant
        if variant not in registry:
            raise NotImplementedError(f"No handler registered for {variant!r}")
        return registry[variant](inputs, **kwargs)

    wrapper.register = register
    update_wrapper(wrapper, func)
    return wrapper


@variant_dispatch
def build_electronic_structure_docs(
    inputs: ESBuilderInput, **kwargs
) -> list[ElectronicStructureDoc]:
    """
    Generate electronic structure documents from tagged input data.

    Dispatches on the variant field of the provided ESBuilderInput to
    construct ElectronicStructureDoc instances via the appropriate
    factory method (from_structure, from_bs, from_dos, or from_bsdos).

    Caller is responsible for creating ESBuilderInput instances
    within their data pipeline context.

    Args:
        inputs: An ESBuilderInput whose variant selects the
            construction path and whose data contains the
            corresponding list of input documents to process.

    Returns:
        list[ElectronicStructureDoc]
    """


@build_electronic_structure_docs.register(Variant.STRUCTURE)
def _(inputs: ESBuilderInput, **kwargs) -> list[ElectronicStructureDoc]:
    return list(
        filter_map(
            ElectronicStructureDoc.from_structure,
            inputs.data,
            work_keys=[
                "deprecated",
                "material_id",
                "meta_structure",
                "origins",
                "warnings",
                "band_gap",
                "cbm",
                "vbm",
                "efermi",
                "is_gap_direct",
                "is_metal",
                "magnetic_ordering",
            ],
            **kwargs,
        )
    )


@build_electronic_structure_docs.register(Variant.BS)
def _(inputs: ESBuilderInput, **kwargs) -> list[ElectronicStructureDoc]:
    return list(
        filter_map(
            ElectronicStructureDoc.from_bs,
            inputs.data,
            work_keys=[
                "bandstructures",
                "origins",
                "structures",
                # PropertyDoc.from_structure(...) kwargs
                "deprecated",
                "material_id",
                "meta_structure",
            ],
            **kwargs,
        )
    )


@build_electronic_structure_docs.register(Variant.DOS)
def _(inputs: ESBuilderInput, **kwargs) -> list[ElectronicStructureDoc]:
    return list(
        filter_map(
            ElectronicStructureDoc.from_dos,
            inputs.data,
            work_keys=[
                "dos",
                "is_gap_direct",
                "origins",
                "structures",
                # PropertyDoc.from_structure(...) kwargs
                "deprecated",
                "material_id",
                "meta_structure",
            ],
            **kwargs,
        )
    )


@build_electronic_structure_docs.register(Variant.BS_DOS)
def _(inputs: ESBuilderInput, **kwargs) -> list[ElectronicStructureDoc]:
    return list(
        filter_map(
            ElectronicStructureDoc.from_bsdos,
            inputs.data,
            work_keys=[
                "bandstructures",
                "dos",
                "origins",
                "structures",
                # PropertyDoc.from_structure(...) kwargs
                "deprecated",
                "material_id",
                "meta_structure",
            ],
            **kwargs,
        )
    )


# -----------------------------------------------------------------------------
# Helper funcs + types
# -----------------------------------------------------------------------------


class BaseCalcInfo(BaseModel):
    """Basic struct of metadata for use in sorting a list of candidate blessed calculations."""

    task_id: str
    is_hubbard: bool | None
    lmaxmix: int | None
    nkpoints: int | None
    last_updated: datetime


class DosCalc(BaseCalcInfo):
    nedos: int | None


class BSCalc(BaseCalcInfo): ...


def obtain_blessed_dos(dos_calcs: list[DosCalc]) -> DosCalc:
    """
    Yields best dos calc from list of dos calcs.

    Helpful for preparing ``ESBuilderInput`` for ``build_electronic_structure_docs``
    """
    sorted_dos_data = sorted(
        dos_calcs,
        key=lambda entry: (
            entry.is_hubbard,
            entry.nkpoints,
            entry.nedos,
            entry.last_updated,
        ),
        reverse=True,
    )
    return sorted_dos_data[0]


def obtain_blessed_bs(bs_calcs: dict[str, list[BSCalc]]) -> dict[str, BSCalc]:
    """
    Yields map of best bs calc per path convention from map of lists of
    bs calcs for each path convention.

    Helpful for preparing ``ESBuilderInput`` for ``build_electronic_structure_docs``
    """
    blessed_entries = {}
    bs_types = ["setyawan_curtarolo", "hinuma", "latimer_munro"]
    for bs_type in bs_types:
        if bs_calcs.get(bs_type):
            sorted_bs_data = sorted(
                [entry for entry in bs_calcs[bs_type] if entry is not None],
                key=lambda entry: (
                    # Entries with any None sort last (False < True, reversed)
                    entry.is_hubbard is not None
                    and entry.nkpoints is not None
                    and entry.last_updated is not None,
                    entry.is_hubbard or False,
                    entry.nkpoints or 0,
                    entry.last_updated or datetime.min,
                ),
                reverse=True,
            )

            if sorted_bs_data:
                blessed_entries[bs_type] = sorted_bs_data[0]

    return blessed_entries



from pydantic import Field

from emmet.core.base import EmmetBaseModel
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.types.typing import IdentifierType


class BaseBuilderInput(EmmetBaseModel):
    """
    Document model with the minimum inputs required
    to run builders that only require a Pymatgen structure
    object for property analysis.

    A material_id and builder_meta information may be optionally
    included.
    """

    deprecated: bool = Field(False)
    material_id: IdentifierType | None = Field(None)
    structure: StructureType



from emmet.builders.base import BaseBuilderInput
from emmet.builders.utils import filter_map
from emmet.core.magnetism import MagnetismDoc
from emmet.core.material import PropertyOrigin


class MagnetismBuilderInput(BaseBuilderInput):
    total_magnetization: float
    origins: list[PropertyOrigin]


def build_magnetism_docs(
    input_documents: list[MagnetismBuilderInput], **kwargs
) -> list[MagnetismDoc]:
    """
    Generate magnetism documents from input structures.

    Transforms a list of MagnetismBuilderInput documents containing
    Pymatgen structures into corresponding MagnetismDoc instances by
    analyzing the magnetic configuration of each structure.

    Caller is responsible for creating MagnetismBuilderInput instances
    within their data pipeline context.

    Args:
        input_documents: List of MagnetismBuilderInput documents to process.

    Returns:
        list[MagnetismDoc]
    """

    return list(
        filter_map(
            MagnetismDoc.from_structure,
            input_documents,
            work_keys=[
                "deprecated",
                "material_id",
                "structure",
                "origins",
                "total_magnetization",
            ],
            **kwargs
        )
    )



import copy
import warnings

from pydantic import BaseModel, Field
from pymatgen.entries.compatibility import Compatibility

from emmet.builders.utils import HiddenPrints
from emmet.core.corrected_entries import CorrectedEntriesDoc
from emmet.core.thermo import ThermoType
from emmet.core.types.pymatgen_types.computed_entries_adapter import (
    ComputedStructureEntryType,
)


class CorrectedEntriesBuilderInput(BaseModel):
    entries: list[ComputedStructureEntryType] = Field(
        ...,
        description="""
        List of computed structure entries to apply corrections to.
        Entries MUST belong to a single chemical system (chemsys).
        """,
    )


def build_corrected_entries_doc(
    input: CorrectedEntriesBuilderInput,
    compatibilities: list[Compatibility | None] = [None],
) -> CorrectedEntriesDoc:
    """
    Process computed structure entries using corrections defined in pymatgen
    compatibility classes. Ensures compatibility of energies for entries for
    different thermodynamic hulls.

    Input entries must all belong to the same chemical system. Caller is
    responsible for constructing CorrectedEntriesBuilderInput instances within
    their data pipeline context.

    Args:
        input: CorrectedEntriesBuilderInput with an aggregated list of computed
            structure entries for a single chemical system.
        compatibilities: List of pymatgen compatibility classes to apply to
            input entries.

    Returns:
        CorrectedEntriesDoc: if no Compatibility class(es) are provided, and all
            entries have the same functional, no corrections will be applied and
            entries will simply be passed through to CorrectedEntriesDoc constructor.
    """
    all_entry_types = {str(e.data["run_type"]) for e in input.entries}

    elements = sorted(
        set([el.symbol for e in input.entries for el in e.composition.elements])
    )
    chemsys = "-".join(elements)

    corrected_entries = {}

    for compatibility in compatibilities:
        if compatibility is not None:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                with HiddenPrints():
                    if compatibility.name == "MP DFT mixing scheme":
                        thermo_type = ThermoType.GGA_GGA_U_R2SCAN

                        if "R2SCAN" in all_entry_types:
                            only_scan_pd_entries = [
                                e
                                for e in input.entries
                                if str(e.data["run_type"]) == "R2SCAN"
                            ]
                            corrected_entries["R2SCAN"] = only_scan_pd_entries

                            pd_entries = compatibility.process_entries(
                                copy.deepcopy(input.entries),
                                verbose=False,
                            )

                        else:
                            corrected_entries["R2SCAN"] = None
                            pd_entries = None

                    elif compatibility.name == "MP2020":
                        thermo_type = ThermoType.GGA_GGA_U
                        pd_entries = compatibility.process_entries(
                            copy.deepcopy(input.entries), verbose=False
                        )
                    else:
                        thermo_type = ThermoType.UNKNOWN
                        pd_entries = compatibility.process_entries(
                            copy.deepcopy(input.entries), verbose=False
                        )

                    corrected_entries[str(thermo_type)] = pd_entries

        else:
            if len(all_entry_types) > 1:
                # TODO: logging over raising
                raise ValueError(
                    "More than one functional type has been provided without a mixing scheme!"
                )
            else:
                thermo_type = all_entry_types.pop()

            corrected_entries[str(thermo_type)] = copy.deepcopy(input.entries)

    return CorrectedEntriesDoc(chemsys=chemsys, entries=corrected_entries)
